'use strict';

// publishing
const {
  app,
  BrowserWindow,
  BrowserView,
  dialog,
  shell: {
    openExternal,
    showItemInFolder,
    openPath,
  },
  ipcMain,
  Menu,
  nativeTheme,
} = require('electron');


// open a node.js supporting window when the app is ready
app.whenReady().then(() => {

  //
  // config: Object<window: Object, save: Function>
  const config = (() => {
    // imports
    const { readFileSync: readFile, writeFileSync: writeFile } = require('fs');

    // prepare the path to the config file
    const path = `${app.getPath('userData')}/config.json`;

    // safely determine the content of the config object
    let config = {};
    try {
      config = JSON.parse(readFile(path, 'utf8'));
    } catch(error) {
      console.error('Error loading config:', error);
      config.window = {};
    }

    // add an asynchronous saving method to the object
    config.save = () => {
      try {
        writeFile(path, JSON.stringify(config, null, 2));
      } catch(error) {
        console.error('Error saving config:', error);
      }
    };

    // return the enriched config object
    return config;
  })();

  // prepare additional options based on the operating system
  const additionalOptions = {};
  switch (process.platform) {
    // optimisations for macOS
    case 'darwin': {
      additionalOptions.titleBarStyle = 'hidden';
      additionalOptions.trafficLightPosition = {
        x: 16,
        y: 16,
      };

      // customise the menubar in order to remove the zoom keys from the view menu
      Menu.setApplicationMenu(Menu.buildFromTemplate([
        {
          role: 'appMenu',
        }, {
          role: 'fileMenu',
        }, {
          role: 'editMenu',
        }, {
          label: 'View',
          submenu: [
            { role: 'reload' },
            { role: 'toggleDevTools' },
            { type: 'separator' },
            { role: 'togglefullscreen' },
          ],
        }, {
          role: 'windowMenu',
        }, {
          role: 'help',
          submenu: [],
        },
      ]));
    } break;

    // optimimisations for windows
    case 'win32': {
      additionalOptions.titleBarOverlay = {
        color: '#00000000',
        height: 48,
      };
      additionalOptions.titleBarStyle = 'hidden';
    } break;
  }

  // open the window
  const win = new BrowserWindow({
    width: config.window.width ?? 1016,
    minWidth: 800,
    height: config.window.height ?? 762,
    minHeight: 600,
    x: config.window.x,
    y: config.window.y,
    autoHideMenuBar: true,
    show: false,
    ...additionalOptions,
    webPreferences: {
      nodeIntegration: true,
      contextIsolation: false,
    },
  });

  // restore the dev tools if the have been opened before, but only after the content of the window
  // has been loaded, as any earlier invocation of the dev tools will crash the app
  if (config.debug) {
    win.webContents.once('did-finish-load', () => win.webContents.openDevTools());
  }

  // detect changes on the dev tools in order to save wether they have been opened before
  ['devtools-opened', 'devtools-closed'].forEach((event) => {
    win.webContents.on(event, () => {
      config.debug = win.webContents.isDevToolsOpened();
      config.save();
    });
  });

  //
  // windowControlColors: Object
  // Colors to use for the window controls on windows.
  const windowControlColors = {
    get system() {
      return this.light;
    },
    light: {
      focus: '#000000',
      blur: '#8b8b8b',
    },
    dark: {
      focus: '#ffffff',
      blur: '#797979',
    },
    update() {
      win.setTitleBarOverlay({
        symbolColor: windowControlColors[nativeTheme.themeSource][win.isFocused() ? 'focus' : 'blur'],
      });
    },
  };

  // use saved theme or use the light mode design by default
  nativeTheme.themeSource = config.theme ?? 'light';

  // update the window control colors right away on windows so they match the restored theme
  if (process.platform === 'win32') {
    windowControlColors.update();
  }

  // listen for changes of the design mode and toggle the whole app instance into light or dark mode
  ipcMain.on('design', (_, design) => {
    // determine the exact theme to set next
    nativeTheme.themeSource = ['light', 'dark'].includes(design) ? design : 'light';

    // on windows we need to update the window controls color
    if (process.platform === 'win32') {
      windowControlColors.update();
    }

    // in case the stored theme differs update it and save the config file
    if (config.theme !== nativeTheme.themeSource) {
      config.theme = nativeTheme.themeSource;
      config.save();
    }
  });

  // ipc handle for toggling the fullscreen mode
  ipcMain.handle('toggleFullScreen', (_, force) => {
    win.setFullScreen(force !== undefined ? force : !win.isFullScreen());
  });

  // ipc handle for accessing system default paths for the given target
  ipcMain.handle('getPath', async (_, target) => {
    try {
      return app.getPath(target);
    } catch {
      return null;
    }
  });

  // ipc handle for showing a file or directory in the native file browser
  ipcMain.handle('showFileOrDirectory', async (_, path) => {
    showItemInFolder(path);
  });

  // ipc handle for opening a file or directory in the associated app
  ipcMain.handle('openFileOrDirectory', async (_, path) => {
    await openPath(path);
  });

  // ipc request for letting the user select a file
  ipcMain.handle('getFilePath', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected path of the user
    return result.filePaths[0];
  });

  // ipc request for letting the user select multiple files
  ipcMain.handle('getFilePaths', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory', 'multiSelections'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected paths of the user
    return result.filePaths;
  });

  // ipc request for letting the user select a directory
  ipcMain.handle('getDirectoryPath', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory', 'openDirectory'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected path of the user
    return result.filePaths[0];
  });

  // ipc request for letting the user select multiple directories
  ipcMain.handle('getDirectoryPaths', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory', 'openDirectory', 'multiSelections'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected paths of the user
    return result.filePaths;
  });

  // ipc request for letting the user select a file or directory
  ipcMain.handle('getFileOrDirectoryPath', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory', 'openDirectory', 'openFile'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected path of the user
    return result.filePaths[0];
  });

  // ipc request for letting the user select multiple files or directories
  ipcMain.handle('getFileOrDirectoryPaths', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showOpenDialog({
      title,
      message,
      properties: ['createDirectory', 'openDirectory', 'openFile', 'multiSelections'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected paths of the user
    return result.filePaths;
  });

  // ipc request for letting the user to save a file
  ipcMain.handle('getSavePath', async ({ title, message } = {}) => {
    // open the dialog and wait for the users input
    const result = await dialog.showSaveDialog({
      title,
      message,
      properties: ['createDirectory'],
    });

    // return nothing if the dialog has been aborted
    if (result.canceled) {
      return null;
    }

    // answer with the selected path of the user
    return result.filePath;
  });

  // create and load the setup view, that checks for code trainer availability. this view is used in
  // order to prevent flickering when the login screen of the code trainer gets loaded.
  const setup = new BrowserView({
    webPreferences: {
      nodeIntegration: true,
      contextIsolation: false,
    },
  });
  setup.webContents.loadFile('./app.html');

  // prepare the setup view
  {
    // add the setup view to the window in the first place
    win.setBrowserView(setup);

    // focus on the content of the setup view to ensure the server selection hotkey works properly
    // in windows as well
    setup.webContents.focus();

    // prepare the initial values of the bounds
    let [ x, y ] = [0, 0];
    let [ width, height ] = win.getContentSize();

    // on macos the y coordinate needs to start with the height of the titlebar
    if (process.platform === 'darwin') {
      y = win.getSize()[1] - height;
    }

    // apply the prepared bound values
    setup.setBounds({ x, y, width, height });

    // ensure the setup view keeps the full size within the window
    setup.setAutoResize({ width: true, height: true });
  }

  // the setup view will trigger the navigation event when the code trainer server shall be loaded,
  // which we interrupt to load the code trainer in the window itself
  setup.webContents.once('will-navigate', (event, url) => {
    // prevent navigation in the view and load the url in the window itself
    event.preventDefault();
    win.loadURL(url);

    // once the windows main content has been loaded remove the setup view for good
    win.webContents.once('did-finish-load', () => {
      // use a timeout to prevent any rarely occuring flickering when removing the setup view
      setTimeout(() => {
        win.removeBrowserView(setup);
        setup.webContents.loadURL('about:blank');
      }, 50);
    });
  });

  // tell the content of the window that it is focused
  win.on('focus', () => {
    // on windows we need to update the window controls color
    if (process.platform === 'win32') {
      windowControlColors.update();
    }

    // set the blur state to the html element of the windows content
    [setup, win].forEach(({ webContents }) => {
      webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-focus', 'yes')`);
    });
  });

  // tell the content of the window that it is blurred
  win.on('blur', () => {
    // on windows we need to update the window controls color
    if (process.platform === 'win32') {
      windowControlColors.update();
    }

    // set the blur state to the html element of the windows content
    [setup, win].forEach(({ webContents }) => {
      webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-focus', 'no')`);
    });
  });

  // on macos we need to communicate to the windows content if we are entering or leaving the full
  // screen mode in order to change the titlebar indent
  if (process.platform === 'darwin') {
    // handle entering the fullscreen mode
    win.on('enter-full-screen', () => {
      [setup, win].forEach(({ webContents }) => {
        webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-fullscreen', 'yes')`);
      });
    });

    // handle leaving the fullscreen mode
    win.on('leave-full-screen', () => {
      [setup, win].forEach(({ webContents }) => {
        webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-fullscreen', 'no')`);
      });
    });
  }

  // store the window bounds when it gets moved or resized
  ['moved', 'resized'].forEach((event) => win.on(event, () => {
    // do not store any updates if we are on macos and in fullscreen mode or
    // windows in maximised window mode, as the app will not start in full-
    // screen mode the next time it gets opened but would still use the full
    // screen size then
    if (
      process.platform === 'darwin' && win.isFullScreen() ||
      process.platform === 'win32' && win.isMaximized()
    ) return;

    // get the current bounds
    const [ width, height ] = win.getSize();
    const [ x, y ] = win.getPosition();

    // update and save the config file with the current properties
    config.window = { width, height, x, y };
    config.save();
  }));

  // apply further customisations to the setup and window contents
  [setup, win].forEach(({ webContents }) => {
    // restore an initial zoom factor in case the user changed it
    webContents.once('did-finish-load', () => {
      webContents.setZoomFactor(1);
    });

    // reapply the window state injections of focus state and fullscreen (macos) whenever the webs
    // contents become ready again (eg. after reload)
    webContents.on('dom-ready', () => {
      // apply the current focus state
      webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-focus', '${win.isFocused() ? 'yes' : 'no'}')`);

      // apply the current fullscreen state on macos
      if (process.platform === 'darwin') {
        webContents.executeJavaScript(`document.documentElement.setAttribute('electron-window-fullscreen', '${win.isFullScreen() ? 'yes' : 'no'}')`);
      }
    });

    // open external links in the browser instead of a new app window
    webContents.setWindowOpenHandler(({ url }) => {
      openExternal(url);
      return { action: 'deny' };
    });
  });

  // show the initially hidden window for the first time when its ready
  setup.webContents.once('did-finish-load', () => {
    // use a timeout here as well in order to prevent any design flickering
    setTimeout(() => {
      win.show();
    }, 50);
  });

});


// close the app altogether when all windows get closed
app.on('window-all-closed', () => {
  app.quit();
});
